from typing import Any, Dict

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateRotate import PrivateRotate
from super_scad.type.Vector3 import Vector3


class Rotate3D(ScadSingleChildParent):
    """
    Rotates its child degrees about the axis of the coordinate system or around an arbitrary axis. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#rotate.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 angle: float | Vector3 | None = None,
                 angle_x: float | None = None,
                 angle_y: float | None = None,
                 angle_z: float | None = None,
                 vector: Vector3 | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param angle: The angle of rotation around all axis or a vector.
        :param angle_x: The angle of rotation around the x-axis.
        :param angle_y: The angle of rotation around the y-axis.
        :param angle_z: The angle of rotation around the z-axis.
        :param vector: The vector of rotation.
        :param child: The widget to be rotated.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._angle: float | Vector3 | None = angle
        """
        The angle of rotation around all axis or a vector.
        """

        self._angle_x: float | None = angle_x
        """
        The angle of rotation around the x-axis.
        """

        self._angle_y: float | None = angle_y
        """
        The angle of rotation around the y-axis.
        """

        self._angle_z: float | None = angle_z
        """
        The angle of rotation around the z-axis.
        """

        self._vector: Vector3 | None = vector
        """ 
        The vector of rotation.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'angle', 'vector'}, {'angle_x', 'angle_y', 'angle_z'})
        validator.validate_required({'angle_x', 'angle_y', 'angle_z', 'angle'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle(self) -> float | Vector3 | None:
        """
        Returns angle of rotation around all axis or a vector.
        """
        if self._vector is not None:
            return self._angle

        if self._angle is None:
            self._angle = Vector3(self.angle_x, self.angle_y, self.angle_z)

        return self._angle

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_x(self) -> float | None:
        """
        Returns the angle of rotation around the x-axis.
        """
        if self._vector is not None:
            return None

        if self._angle_x is None:
            if self._angle is not None:
                self._angle_x = self._angle.x
            else:
                self._angle_x = 0.0

        return self._angle_x

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_y(self) -> float | None:
        """
        Returns the angle of rotation around the y-axis.
        """
        if self._vector is not None:
            return None

        if self._angle_y is None:
            if self._angle is not None:
                self._angle_y = self._angle.y
            else:
                self._angle_y = 0.0

        return self._angle_y

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_z(self) -> float | None:
        """
        Returns the angle of rotation around the z-axis.
        """
        if self._vector is not None:
            return None

        if self._angle_z is None:
            if self._angle is not None:
                self._angle_z = self._angle.z
            else:
                self._angle_z = 0.0

        return self._angle_z

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vector(self) -> Vector3 | None:
        """
        Returns the vector of rotation.
        """
        return self._vector

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateRotate(angle=self.angle, vector=self.vector, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateMirror import PrivateMirror
from super_scad.type.Vector3 import Vector3


class Mirror3D(ScadSingleChildParent):
    """
    Transforms the child widget to a mirror of the original, as if it were the mirror image seen through a plane
    intersecting the origin. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#mirror.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 vector: Vector3 | None = None,
                 x: float | None = None,
                 y: float | None = None,
                 z: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param vector: The normal vector of the origin-intersecting mirror plane used, meaning the vector coming
                       perpendicularly out of the plane. Each coordinate of the original widget is altered such that
                       it becomes equidistant on the other side of this plane from the closest point on the plane.
        :param x: The x-coordinate of the origin-intersecting mirror plane.
        :param y: The y-coordinate of the origin-intersecting mirror plane.
        :param z: The z-coordinate of the origin-intersecting mirror plane.
        :param child: The widget to be mirrored.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._vector: Vector3 | None = vector
        """
        The normal vector of the origin-intersecting mirror plane used, meaning the vector coming perpendicularly out of 
        the plane
        """

        self._x: float | None = x
        """
        The x-coordinate of the origin-intersecting mirror plane.
        """

        self._y: float | None = y
        """
        The y-coordinate of the origin-intersecting mirror plane.
        """

        self._z: float | None = z
        """
        The z-coordinate of the origin-intersecting mirror plane.
        """

        self._normal: Vector3 | None = None
        """
        The normalized normal vector of the origin-intersecting mirror plane used.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'vector'}, {'x', 'y', 'z'})
        validator.validate_required({'x', 'y', 'z', 'vector'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def normal(self) -> Vector3:
        """
        The normal vector of the origin-intersecting mirror plane.
        """
        if self._normal is None:
            if self._vector is not None:
                self._normal = self._vector.normal
            else:
                self._normal = Vector3(self._x or 0.0,
                                       self._y or 0.0,
                                       self._z or 0.0).normal

                self._normal = self._normal * (-1.0 if self._normal.x < 0.0 else 1.0)

        return self._normal

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateMirror(vector=self.normal, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.d3.private.PrivateCube import PrivateCube
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Translate3D import Translate3D
from super_scad.type.Vector3 import Vector3


class Cuboid(ScadWidget):
    """
    Class for cuboids.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 size: Vector3 | None = None,
                 width: float | None = None,
                 depth: float | None = None,
                 height: float | None = None,
                 center: bool = False,
                 extend_by_eps_front: bool = False,
                 extend_by_eps_back: bool = False,
                 extend_by_eps_top: bool = False,
                 extend_by_eps_bottom: bool = False,
                 extend_by_eps_left: bool = False,
                 extend_by_eps_right: bool = False):
        """
        Object constructor.

        :param size: The size of the cuboid.
        :param width: The width (the size along the x-axis) of the cuboid.
        :param depth: The depth (the size along the y-axis) of the cuboid.
        :param height: The height (the size along the y-axis) of the cuboid.
        :param center: Whether the cuboid is centered at the origin.
        :param extend_by_eps_front: Whether to extend the front face of the cuboid by eps for a clear overlap.
        :param extend_by_eps_back: Whether to extend the back face of the cuboid by eps for a clear overlap.
        :param extend_by_eps_top: Whether to extend the top face of the cuboid by eps for a clear overlap.
        :param extend_by_eps_bottom: Whether to extend the bottom face of the cuboid by eps for a clear overlap.
        :param extend_by_eps_left: Whether to extend the left face of the cuboid by eps for a clear overlap.
        :param extend_by_eps_right: Whether to extend the right face of the cuboid by eps for a clear overlap.
        """
        ScadWidget.__init__(self)

        self._size: Vector3 | None = size
        """
        The size of the cuboid.
        """

        self._width: float | None = width
        """
        The width (the size along the x-axis) of the cuboid.
        """

        self._depth: float | None = depth
        """
        The depth (the size along the y-axis) of the cuboid.
        """

        self._height: float | None = height
        """
        The height (the size along the y-axis) of the cuboid.
        """

        self._center: bool = center
        """
        Whether the cuboid is centered at the origin.
        """

        self._extend_by_eps_front: bool = extend_by_eps_front
        """
        Whether to extend the front face of the cuboid by eps for a clear overlap.
        """

        self._extend_by_eps_back: bool = extend_by_eps_back
        """        
        Whether to extend the back face of the cuboid by eps for a clear overlap.
        """

        self._extend_by_eps_top: bool = extend_by_eps_top
        """
        Whether to extend the top face of the cuboid by eps for a clear overlap.
        """

        self._extend_by_eps_bottom: bool = extend_by_eps_bottom
        """
        Whether to extend the bottom face of the cuboid by eps for a clear overlap.
        """

        self._extend_by_eps_left: bool = extend_by_eps_left
        """
        Whether to extend the left face of the cuboid by eps for a clear overlap.
        """

        self._extend_by_eps_right: bool = extend_by_eps_right
        """
        Whether to extend the right face of the cuboid by eps for a clear overlap.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'size'}, {'width', 'depth', 'height'})
        validator.validate_required({'size', 'width'},
                                    {'size', 'depth'},
                                    {'size', 'height'},
                                    {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cuboid is centered at the origin.
        """
        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> Vector3:
        """
        Returns the size of the cuboid.
        """
        if self._size is None:
            self._size = Vector3(x=self.width, y=self.depth, z=self.height)

        return self._size

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def width(self) -> float:
        """
        Returns the width of the cuboid.
        """
        if self._width is None:
            self._width = self._size.x

        return self._width

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def depth(self) -> float:
        """
        Returns the depth of the cuboid.
        """
        if self._depth is None:
            self._depth = self._size.y

        return self._depth

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height of the cuboid.
        """
        if self._height is None:
            self._height = self._size.z

        return self._height

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_front(self) -> bool:
        """
        Return whether to extend the front face of the cuboid by eps for a clear overlap.
        """
        return self._extend_by_eps_front

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_back(self) -> bool:
        """
        Return whether to extend the back face of the cuboid by eps for a clear overlap.
        """
        return self._extend_by_eps_back

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_top(self) -> bool:
        """
        Return whether to extend the top face of the cuboid by eps for a clear overlap.
        """
        return self._extend_by_eps_top

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_bottom(self) -> bool:
        """
        Return whether to extend the bottom face of the cuboid by eps for a clear overlap.
        """
        return self._extend_by_eps_bottom

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_left(self) -> bool:
        """
        Return whether to extend the left face of the cuboid by eps for a clear overlap.
        """
        return self._extend_by_eps_left

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_right(self) -> bool:
        """
        Return whether to extend the front face of the cuboid by eps for a clear overlap.
        """
        return self._extend_by_eps_right

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        cuboid = PrivateCube(size=self._real_size(context), center=self.center)

        translate = self._real_translation(context)
        if translate.is_not_origin:
            cuboid = Translate3D(vector=translate, child=cuboid)

        return cuboid

    # ------------------------------------------------------------------------------------------------------------------
    def _real_translation(self, context) -> Vector3:
        """
        Return the translation of the cuboid accounting for extending by eps.

        @param context: The build context.
        """
        x = 0.0
        y = 0.0
        z = 0.0
        if self.center:
            if self.extend_by_eps_left:
                x -= 1.0
            if self.extend_by_eps_right:
                x += 1.0
            if self.extend_by_eps_front:
                y -= 1.0
            if self.extend_by_eps_back:
                y += 1.0
            if self.extend_by_eps_top:
                z += 1.0
            if self.extend_by_eps_bottom:
                z -= 1.0
            translate = Vector3(x=x, y=y, z=z) * 0.5 * context.eps
        else:
            if self.extend_by_eps_left:
                x -= context.eps
            if self.extend_by_eps_front:
                y -= context.eps
            if self.extend_by_eps_bottom:
                z -= context.eps
            translate = Vector3(x=x, y=y, z=z)

        return translate

    # ------------------------------------------------------------------------------------------------------------------
    def _real_size(self, context) -> Vector3 | float:
        """
        Returns the real size of the cuboid.

        @param context: The build context
        """
        x = 0.0
        y = 0.0
        z = 0.0
        if self.extend_by_eps_left:
            x += context.eps
        if self.extend_by_eps_right:
            x += context.eps
        if self.extend_by_eps_front:
            y += context.eps
        if self.extend_by_eps_back:
            y += context.eps
        if self.extend_by_eps_top:
            z += context.eps
        if self.extend_by_eps_bottom:
            z += context.eps
        size = self.size + Vector3(x=x, y=y, z=z)

        if size.x == size.y and size.x == size.z:
            size = size.x

        return size

# ----------------------------------------------------------------------------------------------------------------------
